/*
    This program shows what can happen in a producer-consumer
    problem if the critical section is not protected.

    The "buffer" is a simple stack type buffer (First-In-Last-Out)
    implemented using an array of char. The buffer starts out half full.
    The "top of stack" (which shows how full the buffer is) is an
    index into the array and it is printed out just below the array as a '^'.
    The producer places '+' characters into the buffer. The consumer takes
    them out leaving '0' characters.

    Since the consumer thread "sleeps" for a longer time than the producer
    thread, the producer is favored in this example, so the buffer starts
    to fill up as the program runs.
*/
#include <windows.h>   // Win32 Sleep()
#include <unistd.h>    // Linux sleep()
#include <stdio.h>
#include "pthread.h"

// global variables
static int runFlag = TRUE;
char theBuffer[80];  // holds one line of screen output
int bufferIndex;
char spaces[79];


int main( int argc, char *argv[])
{
   // producer and consumer thread function prototypes
   void *putIn(void *vargp);
   void *takeOut(void *vargp);

   pthread_t producerID;
   pthread_t consumerID;
   int threadParam1 = 0;  // thread function parameter
   int threadParam2 = 0;

   int i;

   // initialize the spaces array
   for (i = 0; i < 79; i++)
      spaces[i] = ' ';

   // initialize the buffer array
   for (i = 0; i < 40; i++)  // first half of buffer
      theBuffer[i] = '+';
   for (i = 40; i < 80; i++) // second half of buffer
      theBuffer[i] = '0';
   theBuffer[79] = '\0'; // null terminate the buffer
   theBuffer[78] = '*';  // mark the "end" of the buffer
   theBuffer[0] = '*';   // mark the "beginning" the buffer
   // initialize the buffer index to the last '+' in the buffer
   bufferIndex = 39;

   printf("Here's the initial state of the buffer.\n");
   printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);

   // create the producer and consumer threads
   printf("Starting producer and consumer threads\n");
   pthread_create(&producerID, NULL, putIn,   &threadParam1);  // should have error checking
   pthread_create(&consumerID, NULL, takeOut, &threadParam2);

   printf("The producer is thread number %d.\n", producerID);
   printf("The consumer is thread number %d.\n", consumerID);

   Sleep(5000);     // let the threads run for a while
   runFlag = FALSE; // now stop the threads
   Sleep(200);      // give them time to stop
   printf("\nThreads stopped.\n");
   return 0;
}//main


// put a plus sign in the buffer (the producer)
void *putIn(void *vargp)
{
   while(runFlag)
   {
     // begin a critical section
      theBuffer[bufferIndex+1] = '+';
      Sleep(0); // give up the CPU (try commenting this out)
      bufferIndex++;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
     // end of critical section

      Sleep(70); // sleep a short while
   }
   return NULL;
}//putIn


// remove a plus sign from the buffer (the consumer)
void *takeOut(void *vargp)
{
   while(runFlag)
   {
     // begin a critical section
      theBuffer[bufferIndex] = '0';
      Sleep(0); // give up the CPU (try commenting this out)
      bufferIndex--;
      // print the buffer and show the index
      printf("%s\n%.*s^\n", theBuffer, bufferIndex, spaces);
     // end of critical section

      Sleep(120); // sleep a short while
   }
   return NULL;
}//takeOut
